/*      midassound.c
 *
 * HERETIC sound module based on
 *
 * A Sound module for DOOM using MIDAS Digital Audio System. Based on
 * original i_sound.c from the source distribution.
 * 
 * Petteri Kangaslampi, pekangas@sci.fi
*/

//-----------------------------------------------------------------------------
//
// i_sound.c
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This source is available for distribution and/or modification
// only under the terms of the DOOM Source Code License as
// published by id Software. All rights reserved.
//
// The source is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// FITNESS FOR A PARTICULAR PURPOSE. See the DOOM Source Code License
// for more details.
//
// DESCRIPTION:
//	System interface for sound.
//
//-----------------------------------------------------------------------------

const char
midassound_rcsid[] = "$Id: midassound.c,v 1.5 1998/01/07 18:44:18 pekangas Exp $";

// MR (USave)
#ifdef USAVE
#include "USApi.h"
#include "USApiDefs.h"
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include <math.h>

/* We'll need to go below the MIDAS API level a bit */
#include "midas.h"
#include "midasdll.h"

// MR
//#include "z_zone.h"

//#include "i_system.h"
#include "i_sound.h"
//#include "m_argv.h"
//#include "m_misc.h"
//#include "w_wad.h"
//#include "i_video.h"

#include "doomdef.h"

#include "snd_vol.h"

#ifdef DOOM_SND
#define snd_ptr data
#endif

/*#define DEBUGPRINT*/

// The number of internal mixing channels,
//  the samples calculated for each mixing step,
//  the size of the 16bit, 2 hardware channel (stereo)
//  mixing buffer, and the samplerate of the raw data.

#define NUM_CHANNELS		8

#define SAMPLERATE		11025	// Hz
#define SAMPLESIZE		2   	// 16bit

// MR
extern sfxinfo_t S_sfx[];

// The actual lengths of all sound effects.
int 		lengths[NUMSFX];

// The channel data pointers, start and end.
unsigned char*	channels[NUM_CHANNELS];
unsigned char*	channelsend[NUM_CHANNELS];


// Time/gametic that the channel started playing,
//  used to determine oldest, which automatically
//  has lowest priority.
// In case number of active sounds exceeds
//  available channels.
int		channelstart[NUM_CHANNELS];

// The sound in channel handles,
//  determined on registration,
//  might be used to unregister/stop/modify,
//  currently unused.
int 		channelhandles[NUM_CHANNELS];

// SFX id of the playing sound effect.
// Used to catch duplicates (like chainsaw).
int		channelids[NUM_CHANNELS];			

// Pitch to stepping lookup, unused.
int		steptable[256];

// Volume lookups.
int		vol_lookup[128*256];

// Hardware left and right channel volume lookup.
int*		channelleftvol_lookup[NUM_CHANNELS];
int*		channelrightvol_lookup[NUM_CHANNELS];

/* MIDAS sample play handles for all channels: */
MIDASsamplePlayHandle channelPlayHandles[NUM_CHANNELS];

static int wavonly = 0;
static int primarysound = 0;
static int nosound = 0;

// MR
int	snd_MaxVolume,      // maximum volume for sound
	snd_MusicVolume;    // maximum volume for music
int snd_Channels;

// MR
HMIDIOUT	hHmidiDevice=NULL;

tdstSndVolume g_stSfxVol,g_stMidVol;
BOOL g_bSndInitDone=FALSE;

// MR
void I_InitMusic(void);
void I_ShutdownMusic(void);
void I_StopSong(int handle);

void MIDASerror(void)
{
    I_Error("MIDAS Error: %s", MIDASgetErrorMessage(MIDASgetLastError()));
}



/* Loads a sound and adds it to MIDAS - really returns a MIDAS
   sample handle */
void*
getsfx
( char*         sfxname,
  int*          len )
{
    unsigned char*      sfx;
    int                 size;
    char                name[20];
    int                 sfxlump;
    
    int         error;
    static sdSample smp;
    unsigned sampleHandle;

    
    // Get the sound data from the WAD, allocate lump
    //  in zone memory.
    // MR
	//sprintf(name, "ds%s", sfxname);
	sprintf(name, "%s", sfxname);

    // Now, there is a severe problem with the
    //  sound handling, in it is not (yet/anymore)
    //  gamemode aware. That means, sounds from
    //  DOOM II will be requested even with DOOM
    //  shareware.
    // The sound list is wired into sounds.c,
    //  which sets the external variable.
    // I do not do runtime patches to that
    //  variable. Instead, we will use a
    //  default sound for replacement.
    if ( W_CheckNumForName(name) == -1 )
      sfxlump = W_GetNumForName("gldhit");
    else
      sfxlump = W_GetNumForName(name);
    
    size = W_LumpLength( sfxlump );

    // Debug.
    // fprintf( stderr, "." );
    //fprintf( stderr, " -loading  %s (lump %d, %d bytes)\n",
    //	     sfxname, sfxlump, size );
    //fflush( stderr );
    
    sfx = (unsigned char*)W_CacheLumpNum( sfxlump, PU_STATIC );

    /* Add sound to MIDAS: */
    /* A hack below API level - yeahyeah, we should add support for preparing
       samples from memory to the API */
    
    /* Build Sound Device sample structure for the sample: */
    smp.sample = sfx+8;
    smp.samplePos = sdSmpConv;
    smp.sampleType = MIDAS_SAMPLE_8BIT_MONO;
    smp.sampleLength = size-8;
    /* The lump has 8 bytes of magic in the beginning? */

    /* No loop: */
    smp.loopMode = sdLoopNone;
    smp.loop1Start = smp.loop1End = 0;
    smp.loop1Type = loopNone;

    /* No loop 2: */
    smp.loop2Start = smp.loop2End = 0;
    smp.loop2Type = loopNone;

    /* Add the sample to the Sound Device: */
    if ( (error = midasSD->AddSample(&smp, 1, &sampleHandle)) != OK )
        I_Error("getsfx: AddSample failed: %s", MIDASgetErrorMessage(error));

    // Remove the cached lump.
    Z_Free( sfx );
    
    // Preserve padded length.
    *len = size-8;

    /* Return sample handle: (damn ugly) */
    return (void*) sampleHandle;
}





//
// This function adds a sound to the
//  list of currently active sounds,
//  which is maintained as a given number
//  (eight, usually) of internal channels.
// Returns a handle.
//
int
addsfx
( int		sfxid,
  int		volume,
  int		step,
  int		seperation )
{
    static unsigned short	handlenums = 0;
 
    int		i;
    int		rc = -1;
    
    int		oldest = gametic;
    int		oldestnum = 0;
    int		slot;

    MIDASsamplePlayHandle playHandle;
    int vol;
    int pan;
    

    // Chainsaw troubles.
    // Play these sound effects only one at a time.
// MR
    /*if ( sfxid == sfx_sawup
	 || sfxid == sfx_sawidl
	 || sfxid == sfx_sawful
	 || sfxid == sfx_sawhit
	 || sfxid == sfx_stnmov
	 || sfxid == sfx_pistol	 )*/
	/*if ( sfxid == sfx_phopow
		)
    {
	// Loop all channels, check.
	for (i=0 ; i<NUM_CHANNELS ; i++)
	{
	    // Active, and using the same SFX?
	    if ( (channels[i])
		 && (channelids[i] == sfxid) )
	    {
		// Reset.
		channels[i] = 0;
		// We are sure that iff,
		//  there will only be one.
		break;
	    }
	}
    }*/

    // Loop all channels to find oldest SFX.
    for (i=0; (i<NUM_CHANNELS) && (channels[i]); i++)
    {
	if (channelstart[i] < oldest)
	{
	    oldestnum = i;
	    oldest = channelstart[i];
	}
    }

    // Tales from the cryptic.
    // If we found a channel, fine.
    // If not, we simply overwrite the first one, 0.
    // Probably only happens at startup.
    if (i == NUM_CHANNELS)
	slot = oldestnum;
    else
	slot = i;

    // Okay, in the less recent channel,
    //  we will handle the new SFX.
    // Set pointer to raw data.
    channels[slot] = (unsigned char *) S_sfx[sfxid].snd_ptr;
    // Set pointer to end of raw data.
    channelsend[slot] = channels[slot] + lengths[sfxid];

    // Reset current handle number, limited to 0..100.
    if (!handlenums)
	handlenums = 100;

    // Assign current handle number.
    // Preserved so sounds could be stopped (unused).
    channelhandles[slot] = rc = handlenums++;

    // Set stepping???
    // Kinda getting the impression this is never used.
    /*channelstep[slot] = step;*/
    // ???
    /*channelstepremainder[slot] = 0;*/
    // Should be gametic, I presume.
    channelstart[slot] = gametic;

    /* Calculate MIDAS volume and panning from volume and separation: */
    vol = 4*volume;   /* original range 0-15(?), MIDAS range 0-64 */
    pan = (seperation-128) / 2;  /* original 0-255(?), MIDAS -64-64 */

    /* Clamp: */
    if ( vol > 64 ) vol = 64;
    if ( vol < 0 ) vol = 0;
    if ( pan < -64) pan = -64;
    if ( pan > 64) pan = 64;

    // Preserve sound SFX id,
    //  e.g. for avoiding duplicates of chainsaw.
    channelids[slot] = sfxid;

    /* Play the sound: */
    if ( (playHandle = MIDASplaySample((MIDASsample) S_sfx[sfxid].snd_ptr,
                                       slot, 0, SAMPLERATE, vol, pan)) == 0 )
        MIDASerror();

    // You tell me.
/*  [Petteri] We'll return the MIDAS sample playback handle instead
    return rc;*/
    return (int) playHandle;
}





//
// SFX API
// Note: this was called by S_Init.
// However, whatever they did in the
// old DPMS based DOS version, this
// were simply dummies in the Linux
// version.
// See soundserver initdata().
//
void I_SetChannels()
{
  // Init internal lookups (raw data, mixing buffer, channels).
  // This function sets up internal lookups used during
  //  the mixing process. 
  int		i;
  int		j;
    
  int*	steptablemid = steptable + 128;
  
  // Okay, reset internal mixing channels to zero.
  /*for (i=0; i<NUM_CHANNELS; i++)
  {
    channels[i] = 0;
  }*/

  // This table provides step widths for pitch parameters.
  // I fail to see that this is currently used.
  for (i=-128 ; i<128 ; i++)
    steptablemid[i] = (int)(pow(2.0, (i/64.0))*65536.0);
  
  
  // Generates volume lookup tables
  //  which also turn the unsigned samples
  //  into signed samples.
  for (i=0 ; i<128 ; i++)
    for (j=0 ; j<256 ; j++)
      vol_lookup[i*256+j] = (i*(j-128)*256)/127;
}	

// from http://www.codeguru.com/multimedia/volume_manipulation.shtml
#define AUDFREQ			22050	// Frequency
#define AUDCHANNELS		1		// Number of channels
#define AUDBITSMPL		16		// Number of bits per sample
void SetDeviceType( WAVEFORMATEX* pwfe )
{
	WORD  nBlockAlign = (AUDCHANNELS*AUDBITSMPL)/8;
	DWORD nSamplesPerSec = AUDFREQ;
	memset( pwfe, 0, sizeof(WAVEFORMATEX) );
	pwfe->wFormatTag = WAVE_FORMAT_PCM;
	pwfe->nChannels = AUDCHANNELS;
	pwfe->nBlockAlign = nBlockAlign;
	pwfe->nSamplesPerSec = nSamplesPerSec;
	pwfe->wBitsPerSample = AUDBITSMPL;
	pwfe->nAvgBytesPerSec = nSamplesPerSec*nBlockAlign;
}

void I_InitSfxVolume()
{ HWAVEOUT hWavOut;
  WAVEFORMATEX hwfx;
  UINT error;
  MIXERLINE MixerLine;
  MIXERCONTROL Control;
  MIXERLINECONTROLS LineControls;
	
	SetDeviceType( &hwfx );
	error=waveOutOpen(&hWavOut,WAVE_MAPPER,&hwfx,(DWORD)NULL,0,CALLBACK_NULL);

	if (error!=0)	// opening MIDI Output Devices 
	{
		MessageBox(NULL,"The Wave device is used by another application, please close it before running...","MIDI ERROR",MB_ICONEXCLAMATION);
		return;
	}
	
	error = mixerGetID( (HMIXEROBJ)hWavOut, &g_stSfxVol.m_uMixerID, MIXER_OBJECTF_HWAVEOUT );
	if (error!=0)	// opening MIDI Output Devices 
	{
		//MessageBox(NULL,"WaveOut Mapper in Mixer is not available","MIXER ERROR",MB_ICONEXCLAMATION);
		printf("WaveOut Mapper in Mixer is not available\n");
		return;
	}
	
	waveOutClose( hWavOut );

	mixerOpen( (LPHMIXER)&g_stSfxVol.m_dwMixerHandle, g_stSfxVol.m_uMixerID, (DWORD)NULL,0,CALLBACK_NULL );	

	memset( &MixerLine, 0, sizeof(MIXERLINE) );
	MixerLine.cbStruct = sizeof(MIXERLINE);
	MixerLine.dwComponentType = MIXERLINE_COMPONENTTYPE_SRC_WAVEOUT;
	error = mixerGetLineInfo( (HMIXEROBJ)g_stSfxVol.m_dwMixerHandle, &MixerLine, MIXER_GETLINEINFOF_COMPONENTTYPE );
	if ( error != MMSYSERR_NOERROR )
	{
		//MessageBox(NULL,"Could not get WaveOut Destionation Line for the WaveOut Source Line while initilaizing.","MIXER ERROR",MB_ICONEXCLAMATION);
		printf("Could not get WaveOut Destionation Line for the WaveOut Source Line while initilaizing.\n");
		return;
	}

	memset( &Control, 0, sizeof(MIXERCONTROL) );
	Control.cbStruct = sizeof(MIXERCONTROL);

	memset( &LineControls, 0, sizeof(MIXERLINECONTROLS) );
	LineControls.cbStruct = sizeof(MIXERLINECONTROLS);

	LineControls.dwControlType = MIXERCONTROL_CONTROLTYPE_VOLUME;
	LineControls.dwLineID = MixerLine.dwLineID;
	LineControls.cControls = 1;
	LineControls.cbmxctrl = sizeof(MIXERCONTROL);
	LineControls.pamxctrl = &Control;
	error = mixerGetLineControls( (HMIXEROBJ)g_stSfxVol.m_dwMixerHandle, &LineControls, MIXER_GETLINECONTROLSF_ONEBYTYPE );
	if ( error == MMSYSERR_NOERROR )
	{
		if ( !(Control.fdwControl & MIXERCONTROL_CONTROLF_DISABLED) )
		{
//			m_bAvailable = true;
			//MessageBox(NULL,"Volume control for the WaveOut Source Line adopted.","MIXER ERROR",MB_ICONEXCLAMATION);
			//TRACE(".WaveOutputVolume: \"%s\" Volume control for the WaveOut Source Line adopted.\n", Control.szShortName );
		} else {
			MessageBox(NULL,"WaveOutputVolume: WARNING: The Volume Control is disabled.","MIXER ERROR",MB_ICONEXCLAMATION);
			//TRACE(".WaveOutputVolume: WARNING: The Volume Control is disabled.\n" );
		}
	} else {
		MessageBox(NULL,"Could not get the WaveOut Source line Volume Control while initilaizing.","MIXER ERROR",MB_ICONEXCLAMATION);
		//TRACE(".WaveOutputVolume: WARNING: Could not get the WaveOut Source line Volume Control while initilaizing. mmResult=%d\n", mmResult );
	}
	
	g_stSfxVol.m_nChannelCount = MixerLine.cChannels;
	g_stSfxVol.m_dwLineID = LineControls.dwLineID;
	g_stSfxVol.m_dwVolumeControlID = Control.dwControlID;
	g_stSfxVol.m_dwMinimalVolume = Control.Bounds.dwMinimum;
	g_stSfxVol.m_dwMaximalVolume = Control.Bounds.dwMaximum;
//	g_stSfxVol.m_dwVolumeStep = Control.Metrics.cSteps;
}

void I_InitMIDIVolume()
{ //HMIDIOUT hMidiOut;
  UINT error,i;
  MIXERLINE MixerLine;
  MIXERCONTROL Control;
  MIXERLINECONTROLS LineControls;
	
	/*error=midiOutOpen(&hMidiOut,MIDI_MAPPER,(DWORD)NULL,0,CALLBACK_NULL);

	if (error!=0)	// opening MIDI Output Devices 
	{
		MessageBox(NULL,"The Midi device is used by another application, please close it before running...","MIDI ERROR",MB_ICONEXCLAMATION);
		return;
	}
	
	error = mixerGetID( (HMIXEROBJ)hMidiOut, &g_stMidVol.m_uMixerID, MIXER_OBJECTF_HMIDIOUT );*/
	//error=midiOutGetNumDevs();
	for (i=0;i<midiOutGetNumDevs();i++)
	{	error = mixerGetID( (HMIXEROBJ)i, &g_stMidVol.m_uMixerID, MIXER_OBJECTF_MIDIOUT );
		if (!error)
			break;
	}
	if (error)	// opening MIDI Output Devices 
	{
		//MessageBox(NULL,"MidiOut Mapper in Mixer is not available","MIXER ERROR",MB_ICONEXCLAMATION);
		printf("MidiOut Mapper in Mixer is not available\n");
//		midiOutClose( hMidiOut );
		return;
	}
	
//	midiOutClose( hMidiOut );

	mixerOpen( (LPHMIXER)&g_stMidVol.m_dwMixerHandle, g_stMidVol.m_uMixerID, (DWORD)NULL,0,CALLBACK_NULL );	

	memset( &MixerLine, 0, sizeof(MIXERLINE) );
	MixerLine.cbStruct = sizeof(MIXERLINE);
	MixerLine.dwComponentType = MIXERLINE_COMPONENTTYPE_SRC_SYNTHESIZER;
	error = mixerGetLineInfo( (HMIXEROBJ)g_stMidVol.m_dwMixerHandle, &MixerLine, MIXER_GETLINEINFOF_COMPONENTTYPE );
	if ( error != MMSYSERR_NOERROR )
	{
		//MessageBox(NULL,"Could not get MidiOut Destionation Line for the WaveOut Source Line while initilaizing.","MIXER ERROR",MB_ICONEXCLAMATION);
		printf("Could not get MidiOut Destionation Line for the WaveOut Source Line while initilaizing.\n");
		return;
	}

	memset( &Control, 0, sizeof(MIXERCONTROL) );
	Control.cbStruct = sizeof(MIXERCONTROL);

	memset( &LineControls, 0, sizeof(MIXERLINECONTROLS) );
	LineControls.cbStruct = sizeof(MIXERLINECONTROLS);

	LineControls.dwControlType = MIXERCONTROL_CONTROLTYPE_VOLUME;
	LineControls.dwLineID = MixerLine.dwLineID;
	LineControls.cControls = 1;
	LineControls.cbmxctrl = sizeof(MIXERCONTROL);
	LineControls.pamxctrl = &Control;
	error = mixerGetLineControls( (HMIXEROBJ)g_stMidVol.m_dwMixerHandle, &LineControls, MIXER_GETLINECONTROLSF_ONEBYTYPE );
	if ( error == MMSYSERR_NOERROR )
	{
		if ( !(Control.fdwControl & MIXERCONTROL_CONTROLF_DISABLED) )
		{
//			m_bAvailable = true;
			//MessageBox(NULL,"Volume control for the WaveOut Source Line adopted.","MIXER ERROR",MB_ICONEXCLAMATION);
			//TRACE(".WaveOutputVolume: \"%s\" Volume control for the WaveOut Source Line adopted.\n", Control.szShortName );
		} else {
			//MessageBox(NULL,"MidiOutputVolume: WARNING: The Volume Control is disabled.","MIXER ERROR",MB_ICONEXCLAMATION);
			printf("MidiOutputVolume: WARNING: The Volume Control is disabled.\n");
			//TRACE(".WaveOutputVolume: WARNING: The Volume Control is disabled.\n" );
		}
	} else {
		MessageBox(NULL,"Could not get the MidiOut Source line Volume Control while initilaizing.","MIXER ERROR",MB_ICONEXCLAMATION);
		//TRACE(".WaveOutputVolume: WARNING: Could not get the WaveOut Source line Volume Control while initilaizing. mmResult=%d\n", mmResult );
	}
	
	g_stMidVol.m_nChannelCount = MixerLine.cChannels;
	g_stMidVol.m_dwLineID = LineControls.dwLineID;
	g_stMidVol.m_dwVolumeControlID = Control.dwControlID;
	g_stMidVol.m_dwMinimalVolume = Control.Bounds.dwMinimum;
	g_stMidVol.m_dwMaximalVolume = Control.Bounds.dwMaximum;
//	g_stSfxVol.m_dwVolumeStep = Control.Metrics.cSteps;
}

void I_SetSfxVolume(int volume)
{ /*HWAVEOUT hWavOut;
  WAVEFORMATEX hwfx;*/
  UINT error;
  int i;
  DWORD dwVolume;
  MIXERCONTROLDETAILS ControlDetails;
  MIXERCONTROLDETAILS_UNSIGNED* aDetails = (MIXERCONTROLDETAILS_UNSIGNED*)malloc(g_stSfxVol.m_nChannelCount*sizeof(MIXERCONTROLDETAILS_UNSIGNED));

  // Identical to DOS.
  // Basically, this should propagate
  //  the menu/config file setting
  //  to the state variable used in
  //  the mixing.
  snd_MaxVolume = volume;

	/*error=waveOutOpen(&hWavOut,WAVE_MAPPER,&hwfx,(DWORD)NULL,0,CALLBACK_NULL);

	if (error!=0)	// opening MIDI Output Devices 
	{
		MessageBox(NULL,"The Wave device is used by another application, please close it before running...","MIDI ERROR",MB_ICONEXCLAMATION);
		return;
	}*/

  //volume*=16;
  //dwVolume=(((DWORD)volume&0xff)<<8)+(volume&0xff);
  /*waveOutSetVolume(hHmidiDevice,dwVolume);
	
  waveOutClose(hHmidiDevice);*/
  dwVolume=volume*g_stSfxVol.m_dwMaximalVolume/16;
	if ( (dwVolume<g_stSfxVol.m_dwMinimalVolume) || (dwVolume>g_stSfxVol.m_dwMaximalVolume) )
		return;
	
	if ( !aDetails )
		return;
	for ( i = 0; i < g_stSfxVol.m_nChannelCount; i++ )
	{
		aDetails[i].dwValue = dwVolume;
	}
	
	memset( &ControlDetails, 0, sizeof(MIXERCONTROLDETAILS) );
	ControlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
	ControlDetails.dwControlID = g_stSfxVol.m_dwVolumeControlID;
	ControlDetails.cChannels = g_stSfxVol.m_nChannelCount;
	ControlDetails.cMultipleItems = 0;
	ControlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
	ControlDetails.paDetails = &aDetails[0];
	error = mixerSetControlDetails( (HMIXEROBJ)g_stSfxVol.m_dwMixerHandle, &ControlDetails, MIXER_SETCONTROLDETAILSF_VALUE );
	free( aDetails );
	if ( error != MMSYSERR_NOERROR )
	{
		MessageBox(NULL,"Could not set volume","WAVE ERROR",MB_ICONEXCLAMATION);
		//TRACE(".WaveOutputVolume: FAILURE: Could not set volume(%d) mmResult=%d\n", dwValue, mmResult );
	}

}

// MUSIC API - dummy. Some code from DOS version.
void I_SetMusicVolume(int volume)
{ DWORD dwVolume;
  UINT error;
  int i;
  MIXERCONTROLDETAILS ControlDetails;
  MIXERCONTROLDETAILS_UNSIGNED* aDetails = (MIXERCONTROLDETAILS_UNSIGNED*)malloc(g_stMidVol.m_nChannelCount*sizeof(MIXERCONTROLDETAILS_UNSIGNED));
  
  // Internal state variable.
  snd_MusicVolume = volume;
  // Now set volume on output device.

  dwVolume=volume*g_stMidVol.m_dwMaximalVolume/16;
  //dwVolume=(((DWORD)volume&0xffff)<<16)+(volume&0xffff);
	if ( (dwVolume<g_stMidVol.m_dwMinimalVolume) || (dwVolume>g_stMidVol.m_dwMaximalVolume) )
		return;
	
	if ( !aDetails )
		return;
	for ( i = 0; i < g_stMidVol.m_nChannelCount; i++ )
	{
		aDetails[i].dwValue = dwVolume;
	}
	
	memset( &ControlDetails, 0, sizeof(MIXERCONTROLDETAILS) );
	ControlDetails.cbStruct = sizeof(MIXERCONTROLDETAILS);
	ControlDetails.dwControlID = g_stMidVol.m_dwVolumeControlID;
	ControlDetails.cChannels = g_stMidVol.m_nChannelCount;
	ControlDetails.cMultipleItems = 0;
	ControlDetails.cbDetails = sizeof(MIXERCONTROLDETAILS_UNSIGNED);
	ControlDetails.paDetails = &aDetails[0];
	error = mixerSetControlDetails( (HMIXEROBJ)g_stMidVol.m_dwMixerHandle, &ControlDetails, MIXER_SETCONTROLDETAILSF_VALUE );
	free( aDetails );
	if ( error != MMSYSERR_NOERROR )
	{
		printf("Could not set volume\n");
		//TRACE(".WaveOutputVolume: FAILURE: Could not set volume(%d) mmResult=%d\n", dwValue, mmResult );
	}
}
void I_SetMusicVolume2(int volume)
{ DWORD dwVolume;
  UINT error;
  
  // Internal state variable.
  snd_MusicVolume = volume;
  // Now set volume on output device.
	error=midiOutOpen(&hHmidiDevice,MIDI_MAPPER,(DWORD)NULL,0,CALLBACK_NULL);

	if (error!=0)	// opening MIDI Output Devices 
	{
		MessageBox(NULL,"The Midi device is used by another application, please close it before running...","MIDI ERROR",MB_ICONEXCLAMATION);
		return;
	}

  volume*=4096;
  dwVolume=(((DWORD)volume&0xffff)<<16)+(volume&0xffff);
  midiOutSetVolume(hHmidiDevice,dwVolume);

  midiOutClose(hHmidiDevice);
}


//
// Retrieve the raw data lump index
//  for a given SFX name.
//
int I_GetSfxLumpNum(sfxinfo_t* sfx)
{
// MR
/*    char namebuf[9];
    sprintf(namebuf, "ds%s", sfx->name);
    return W_GetNumForName(namebuf);*/
	// MR (from I_SOUND.C)

  if(sfx->name == 0)
		return 0;
  if (sfx->link) sfx = sfx->link;
//  sprintf(namebuf, "d%c%s", snd_prefixen[snd_SfxDevice], sound->name);
  return W_GetNumForName(sfx->name);

}

//
// Starting a sound means adding it
//  to the current list of active sounds
//  in the internal channels.
// As the SFX info struct contains
//  e.g. a pointer to the raw data,
//  it is ignored.
// As our sound handling does not handle
//  priority, it is ignored.
// Pitching (that is, increased speed of playback)
//  is set, but currently not used by mixing.
//
int
I_StartSound
( int		id,
  int		vol,
  int		sep,
  int		pitch,
  int		priority )
{

  // UNUSED
  priority = 0;

#ifdef DEBUGPRINT  
  fprintf(stderr, "I_StartSound: vol %i, sep %i, pitch %i, pri %i\n",
          vol, sep, pitch, priority);
#endif  
  
    // Returns a handle (not used).
    id = addsfx( id, vol, steptable[pitch], sep );

    return id;
}



void I_StopSound (int handle)
{
  // You need the handle returned by StartSound.
  // Would be looping all channels,
  //  tracking down the handle,
  //  an setting the channel to zero.

#ifdef DEBUGPRINT    
    fprintf(stderr, "I_StopSound\n");
#endif    
    /* [Petteri] We'll simply stop the sound with MIDAS (?): */
    if ( !MIDASstopSample((MIDASsamplePlayHandle) handle) )
        MIDASerror();
}


int I_SoundIsPlaying(int handle)
{
    // Ouch.
/*    return gametic < handle;*/
    int is;
    is = (int) MIDASgetSamplePlayStatus((MIDASsamplePlayHandle) handle);
#ifdef DEBUGPRINT    
    fprintf(stderr, "I_SoundIsPlaying: %i\n", is);
#endif    
    return is;
}



int I_QrySongPlaying(int handle);


void I_UpdateSound( void )
{
    /* Loop song if necessary: */
    I_QrySongPlaying(1);
}


// 
// This would be used to write out the mixbuffer
//  during each game loop update.
// Updates sound buffer and audio device at runtime. 
// It is called during Timer interrupt with SNDINTR.
// Mixing now done synchronous, and
//  only output be done asynchronous?
//
void
I_SubmitSound(void)
{
  // Write it to DSP device.
/*  write(audio_fd, mixbuffer, SAMPLECOUNT*BUFMUL);*/
}



void
I_UpdateSoundParams
( int	handle,
  int	vol,
  int	sep,
  int	pitch)
{
    int mvol, mpan;

    pitch = pitch; /* how to use? */
    
    /* Calculate MIDAS volume and panning from volume and separation: */
    mvol = 4*vol;   /* original range 0-15(?), MIDAS range 0-64 */
    mpan = (sep-128) / 2;  /* original 0-255(?), MIDAS -64-64 */

    /* Clamp: */
    if ( mvol > 64 ) mvol = 64;
    if ( mvol < 0 ) mvol = 0;
    if ( mpan < -64) mpan = -64;
    if ( mpan > 64) mpan = 64;

    /* Set: */
    if ( !MIDASsetSampleVolume((MIDASsamplePlayHandle) handle, mvol) )
        MIDASerror();
    if ( !MIDASsetSamplePanning((MIDASsamplePlayHandle) handle, mpan) )
        MIDASerror();
}




void I_ShutdownSound(void)
{
    unsigned i;

	if (!g_bSndInitDone)
		return;
    fprintf(stderr, "I_ShutdownSound: Stopping sounds\n");
    for ( i = 0; i < NUM_CHANNELS; i++ )
        MIDASstopSample(channelPlayHandles[i]);
    
    fprintf(stderr, "I_ShutdownSound: Uninitializing MIDAS\n");
    if ( !MIDASstopBackgroundPlay() )
        MIDASerror();
    if ( !MIDAScloseChannels() )
        MIDASerror();
    if ( !MIDASclose() )
        MIDASerror();

    I_ShutdownMusic();
}




extern DWORD win;  /* window handle from win32video.c (actually HWND) */


void
I_InitSound()
{
    unsigned i;

    I_InitSfxVolume();	// MR
	I_InitMIDIVolume();

	I_InitMusic();

    /* Get command line options: */
    wavonly = !!M_CheckParm("-wavonly");
    primarysound = !!M_CheckParm("-primarysound");
    nosound = !!M_CheckParm("-nosound");
    
    fprintf(stderr, "I_InitSound: Initializing MIDAS\n");
    
    MIDASstartup();
    
    MIDASsetOption(MIDAS_OPTION_MIXRATE, SAMPLERATE);
    MIDASsetOption(MIDAS_OPTION_MIXBUFLEN, 200);

    /* We'll need to initialize graphics here so that we can get a window
       handle for DirectSound: */
    I_InitGraphics();
    
    if ( !wavonly )
    {
        MIDASsetOption(MIDAS_OPTION_DSOUND_HWND, win);
        if ( primarysound )
            MIDASsetOption(MIDAS_OPTION_DSOUND_MODE, MIDAS_DSOUND_PRIMARY);
        else
            MIDASsetOption(MIDAS_OPTION_DSOUND_MODE, MIDAS_DSOUND_STREAM);
    }

    if ( nosound )
        MIDASsetOption(MIDAS_OPTION_FORCE_NO_SOUND, TRUE);  
    
    if ( !MIDASinit() )
        MIDASerror();
    if ( !MIDASstartBackgroundPlay(100) )
        MIDASerror();
    if ( !MIDASopenChannels(NUM_CHANNELS) )
        MIDASerror();

    for ( i = 0; i < NUM_CHANNELS; i++ )
        channelPlayHandles[i] = 0;

    /* Simply preload all sounds and throw them at MIDAS: */
    fprintf(stderr, "I_InitSound: Loading all sounds\n");
    for ( i = 1; i < NUMSFX; i++ )
    {
        if ( !S_sfx[i].link )
        {
            /* Fine, not an alias, just load it */
            S_sfx[i].snd_ptr = getsfx(S_sfx[i].name, &lengths[i]);
            /* getsfx() actually returns a MIDAS sample handle */
        }
/*        else
        {
            // An alias sound
            S_sfx[i].snd_ptr = S_sfx[i].link->snd_ptr;
            lengths[i] = lengths[(S_sfx[i].link - S_sfx)/sizeof(sfxinfo_t)];
        }*/
    }
    for ( i = 1; i < NUMSFX; i++ )	// MR2905
    {
        if ( S_sfx[i].link )
        {
            // An alias sound
            S_sfx[i].snd_ptr = S_sfx[i].link->snd_ptr;
            lengths[i] = lengths[(S_sfx[i].link - S_sfx)/sizeof(sfxinfo_t)];
        }
    }

  // Finished initialization.
  fprintf(stderr, "I_InitSound: sound module ready\n");  
  g_bSndInitDone=TRUE;
}




/* Music using Windows MIDI device - yuck */

static int nomusic = 0;
static char musName[512];
static char midName[512];
static int	looping=0;
static int	musicdies=-1;
static int playing = 0;
static int muslen;


void I_InitMusic(void)
{
    char *temp;

    fprintf(stderr, "I_InitMusic\n");
    
    nomusic = !!M_CheckParm("-nomusic");

    /* Create temporary file names: */    
    temp = getenv("TEMP");
    if ( temp == NULL )
        temp = ".";
    strcpy(musName, temp);

    while ( musName[strlen(musName)-1] == '\\' )
        musName[strlen(musName)-1] = 0;

    strcat(musName, "\\");
    strcpy(midName, musName);
    strcat(musName, "doomtemp.mus");
    strcat(midName, "doomtemp.mid");
}


void I_ShutdownMusic(void)
{
    if ( playing )
        I_StopSong(1);
    
    remove(musName);
    remove(midName);
}



int SendMCI(char *str, char *retStr, int retLen)
{
    int res;
    char errorStr[256];

    res = mciSendString(str, retStr, retLen, NULL);
    if ( res )
    {
        mciGetErrorString(res, errorStr, 255);
        fprintf(stderr, "MCI error: %s\n", errorStr);
    }
    
    return res;
}


void I_PlaySong(int handle, int _looping)
{
    char command[768];
    char resp[32];
        
    if ( nomusic || (handle == -1) )
    {
        playing = 0;
        return;
    }

    handle = handle;
    looping = _looping;
    playing = 0;

    sprintf(command, "open %s alias doomMusic", midName);
    if ( SendMCI(command, NULL, 0) )
        return;
    
    if ( SendMCI("play doomMusic from 0", NULL, 0) )
        return;

    if ( SendMCI("set doomMusic time format ms", NULL, 0) )
        return;

    if ( SendMCI("status doomMusic length", resp, 31) )
        return;

    muslen = atoi(resp);

    musicdies = gametic + (muslen * TICRATE) / 1000;

    playing = 1;
}

void I_PauseSong (int handle)
{
    if ( !playing )
        return;
    
    handle = handle;
}

void I_ResumeSong (int handle)
{
    if ( !playing )
        return;
    
    handle = handle;
}

void I_StopSong(int handle)
{
    handle = handle;
    
    if ( !playing )
        return;

    playing = 0;

    SendMCI("close doomMusic", NULL, 0);
}

void I_UnRegisterSong(int handle)
{
    handle = handle;
    remove(midName);
}

int musicLump;
extern int convert( const char *mus, const char *mid, int nodisplay, int div,
                    int size, int nocomp, int *ow );


int I_RegisterSong(void* data)
{
    FILE *f;
    int ow = 2;

    if ( (f = fopen(musName, "wb")) == NULL )
        I_Error("Unable to open temporary MUS file %s", musName);
    if ( fwrite(data, W_LumpLength(musicLump), 1, f) != 1 )
        I_Error("Unable to write temporary MUS file");
    fclose(f);

    convert(musName, midName, 1, 89, 0, 1, &ow);

    remove(musName);
  
    return 1;
}

// Is the song playing?
int I_QrySongPlaying(int handle)
{
    handle = handle;
    if ( !playing )
        return 0;

    if ( looping )
    {
        if ( gametic >= musicdies )
        {
            SendMCI("play doomMusic from 0", NULL, 0);
            musicdies = gametic + (muslen * TICRATE) / 1000;
        }
        return 1;
    }

    return musicdies > gametic;
}

